{-|
Module      : Gargantext.Core.Text.DeclarativeNgrams
Description : "Declarative" ngrams, i.e. plain text with optional modifiers
Copyright   : (c) CNRS, 2017-Present
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX

See https://gitlab.iscpif.fr/gargantext/haskell-gargantext/issues/386#note_12413
-}

{-# OPTIONS_GHC -Wno-deprecations #-}  -- warning about 'undefined' usage is thrown

module Gargantext.Core.Text.DeclarativeNgrams where


import Data.Text qualified as T
import Gargantext.Core (Lang(..))
import Gargantext.Prelude



data Ngrams =
  Ngrams { terms :: Text
         , declaration :: Declaration }
  deriving (Eq, Show)

data Declaration =
    Exact
  | CaseInsensitive
  | SingularPlural Lang
  -- TODO: SingularPlural Lang
  -- TODO: FrenchDeterminers (https://en.wikipedia.org/wiki/French_articles_and_determiners)
  -- TODO Regex
  -- TODO Combine CaseInsensitive and SingluarPlural
  -- TODO Combine CaseInsensitive and FrenchDeterminers
  deriving (Eq, Show)


data NgramsWithPos =
  NgramsWithPos { ngrams :: Ngrams
                , match  :: Text
                , pos    :: Int }
  deriving (Eq, Show)




-- | TODO Use parsec for this?
-- e.g. matching case-insensitive:
-- https://stackoverflow.com/questions/12937325/whats-the-cleanest-way-to-do-case-insensitive-parsing-with-text-combinators-par
findNgrams :: [Ngrams] -> Text -> [NgramsWithPos]
findNgrams [] _txt = []
findNgrams (ng:ngs) txt = findNgrams' ng txt 0 <> (findNgrams ngs txt)

findNgrams' :: Ngrams -> Text -> Int -> [NgramsWithPos]
findNgrams' _ "" _ = []
findNgrams' ngrams@(Ngrams { terms, declaration }) txt initialPos =
  foldl (<>) [] $ map handleBreak (breakOnDeclaration declaration terms txt)
  where
    handleBreak :: (Text, Text, Text) -> [NgramsWithPos]
    handleBreak (s, matching, _e) =
      let pos = initialPos + T.length s
          txt' = T.drop (T.length s) txt
          match = T.take (T.length matching) txt'
          ngramsWithPos = NgramsWithPos { ngrams, pos, match } in
      (ngramsWithPos : (findNgrams' ngrams (T.drop (T.length matching) txt') (pos + T.length matching)))

-- | Split declaration function in spirit of `T.breakOn` + Declaration type
breakOnDeclaration :: Declaration -> Text -> Text -> [(Text, Text, Text)]
breakOnDeclaration Exact terms txt =
  case T.breakOn terms txt of
    (_, "") -> []
    (s, e) -> [(s, terms, T.drop (T.length terms) e)]
breakOnDeclaration CaseInsensitive terms txt =
  case T.breakOn (T.toLower terms) (T.toLower txt) of
    (_, "") -> []
    (s, e) -> [(s, T.take (T.length terms) e, T.drop (T.length terms) e)]
breakOnDeclaration (SingularPlural lang) terms txt = singularOrPlural
  where
    pf = pluralForm lang terms
    -- | Singular is a bit tricky, because it's a "subset" of plural,
    -- without the "s", so we filter things out.
    -- singular = filter (\(_s, m, e) -> T.take (T.length pf) (m <> e) /= pf) $ breakOnDeclaration Exact terms txt
    singular = breakOnDeclaration Exact terms txt
    plural = breakOnDeclaration Exact pf txt

    singularOrPlural = case (singular, plural) of
      ([s], [p]) ->
        let (ss, _, _) = s
            (sp, _, _) = p in
        if T.length ss < T.length sp then [s] else [p]
      ([s], _) -> [s]
      (_, p) -> p

pluralForm :: Lang -> Text -> Text
pluralForm EN "door" = "door"
pluralForm EN "sheep" = "sheep"
pluralForm EN term = term <> "s"
pluralForm _ _ = undefined
